from abc import abstractmethod

from telegram.bot import Bot


class MessengerError(Exception):
    pass


class Messenger:

    @abstractmethod
    def send(self, content: str) -> None:
        pass


class TelegramMessenger(Messenger):

    def __init__(self, token, chat_id):
        self.token = token
        self.chat_id = chat_id
        self.bot = Bot(self.token)

    def send(self, content: str) -> None:
        try:
            self.bot.send_message(chat_id=self.chat_id, text=content)
        except Exception as error:
            raise MessengerError("Unexpected error occurred at messenger send()") from error



from typing import List, Optional, Union
from datetime import datetime

from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.executors.pool import ThreadPoolExecutor

from .messenger import Messenger, MessengerError
from .scraper import Article, ArticleScraper, validate_scraper
from .press import validate_press_names, Press
from .keywords import Keyword, drop_duplicated_keywords
from .container import ArticleContainer, QueueContainer


class ScraperRunner:

    def __init__(
        self,
        messenger: Messenger,
        scrapers: Optional[List[ArticleScraper]] = None,
        press: Optional[List[Press]] = None,
        keywords: Optional[List[Keyword]] = None,
        article_container: Optional[ArticleContainer] = None,
    ):
        self.messenger = messenger
        self.scheduler = None
        self.scrapers = scrapers if scrapers else []
        self.press = press if press else []
        self.keywords = keywords if keywords else []
        self.article_container = article_container if article_container else QueueContainer()

    def _get_scheduler(self):
        executors = {
            "default": ThreadPoolExecutor(len(self.scrapers)),
        }
        return BlockingScheduler(executors=executors)

    def _set_stop_signal_handler(self):
        # pylint: disable=import-outside-toplevel
        import signal
        signal.signal(signal.SIGINT, self.stop)

    def get_scraper_press_names(self) -> List[str]:
        return [each.get_press_name() for each in self.scrapers]

    def validate_messenger(self) -> None:
        try:
            self.messenger.send("Validation message from nscrap")
            print("[+] Send validation message from nscrap")
        except MessengerError as error:
            print(error)
            print("[+] Messenger went wrong, check messenger.send()")
            raise

    def add_press(self, press: Union[Press, List[Press]]) -> None:
        if not isinstance(press, list):
            self.press.append(press)
        else:
            self.press.extend(press)

    def add_keyword(self, keywords: Union[Keyword, List[Keyword]]) -> None:
        if not isinstance(keywords, list):
            self.keywords.append(keywords)
        else:
            self.keywords.extend(keywords)
        self.keywords = drop_duplicated_keywords(self.keywords)

    def add_scraper(self, scrapers: Union[ArticleScraper, List[ArticleScraper]]) -> None:
        if not isinstance(scrapers, list):
            self.scrapers.append(scrapers)
        else:
            self.scrapers.extend(scrapers)

    def send_article_message(self, article: Article) -> None:
        content = article.to_mesage_format()
        print(f"[+] Send message {content}")
        self.messenger.send(content)

    def is_contains_any_keyword(self, word) -> bool:
        return any([keyword.is_in(word) for keyword in self.keywords])

    def get_new_articles(self, scraped_articles: List[Article]) -> List[Article]:
        existing_article_titles = [each.title for each in self.article_container.get_all_articles()]
        if not scraped_articles:
            return []
        return list(filter(lambda x: x.title not in existing_article_titles, scraped_articles))

    def scrap_new_articles(self, scraper: ArticleScraper) -> List[Article]:
        print(f"[+] Start {scraper.get_press_name()} scraper at {datetime.now():%Y-%m-%d %H:%M:%S}")
        articles = scraper.get_articles()
        return self.get_new_articles(articles)

    def scrap(self, scraper: ArticleScraper) -> None:
        try:
            new_articles = self.scrap_new_articles(scraper)
            for article in new_articles:
                print(f"[+] Scrap {article.to_mesage_format()}")
                if self.is_contains_any_keyword(article.title):
                    self.send_article_message(article)
            self.article_container.add_articles(new_articles)
        except MessengerError:
            # messenger api limit error
            print("[+] MessengerError occurred. Messenger api limit may have been exceeded")

    def print_settings(self) -> None:
        print(f"press: {self.press}")
        print(f"keywords: {[each.keyword for each in self.keywords]}")
        print(f"scrapers: {[each.get_press_name() for each in self.scrapers]}")

    def _add_scheduler_jobs(self):
        scraper_press_names = self.get_scraper_press_names()
        for each in self.press:
            if not each.active:
                continue
            index = scraper_press_names.index(each.press_name)
            scraper = self.scrapers[index]
            scraper_press_name = scraper.get_press_name()
            if not validate_scraper(scraper):
                print(f"[+] Inactive {scraper_press_name} scraper because an error has occurred")
                continue
            self.scheduler.add_job(
                self.scrap,
                "interval",
                args=[scraper],
                seconds=each.delay
            )

    def _start_scheduler(self):
        print("[+] Start nscrap")
        self.scheduler.start()

    def _stop_scheduler(self):
        print("[+] Stop nscrap")
        self.scheduler.shutdown()

    def start(self) -> None:
        validate_press_names(self.press, self.get_scraper_press_names())
        self.scheduler = self._get_scheduler()
        self.validate_messenger()
        self._set_stop_signal_handler()
        self._add_scheduler_jobs()
        self.print_settings()
        self._start_scheduler()

    def stop(self, sig, frame) -> None:
        # pylint: disable=unused-argument
        self._stop_scheduler()



from abc import abstractmethod
from typing import List, Optional
from dataclasses import dataclass
from datetime import datetime


@dataclass
class Article:

    title: str
    link: str
    timestamp: datetime
    content: Optional[str] = None

    def to_mesage_format(self):
        return f"{self.title}({self.link})"


class ArticleScraper:

    @abstractmethod
    def get_press_name(self) -> str:
        pass

    @abstractmethod
    def get_articles(self) -> List[Article]:
        pass


class ArticleConnectionError(Exception):
    pass


class ArticleParsingError(Exception):
    pass


def validate_scraper(scraper: ArticleScraper) -> bool:
    scraper_press_name = scraper.get_press_name()
    try:
        scraper.get_articles()
        print(f"[+] Test succeed: {scraper_press_name} passed test")
        return True
    except ArticleConnectionError:
        print(f"[+] Test failed: {scraper_press_name} scraper could not connect to web site")
        return False
    except ArticleParsingError:
        # pylint: disable=line-too-long
        print(f"[+] Test failed: {scraper_press_name} scraper could not parse web page. Web page may have been changed.")
        return False
    except Exception as error:  # pylint: disable=broad-except
        print(error)
        print(f"[+] Test failed: Unhandled error occurred in {scraper_press_name} scraper")
        return False



from dataclasses import dataclass
from typing import List


@dataclass
class Press:

    press_name: str
    active: bool
    delay: int


def validate_press_names(press: List[Press], scraper_press_names: List[str]):
    press_names = [each.press_name for each in press]
    if set(scraper_press_names) > set(press_names):
        raise ValueError(f"Wrong press names: {press_names}")



from abc import abstractmethod
from typing import List

from .scraper import Article


class ArticleContainer:

    @abstractmethod
    def get_all_articles(self) -> List[Article]:
        pass

    @abstractmethod
    def add_articles(self, articles: List[Article]) -> None:
        pass


class QueueContainer(ArticleContainer):

    def __init__(self):
        # pylint: disable=import-outside-toplevel
        from queue import Queue
        self._article_container = Queue()

    def get_all_articles(self):
        return list(self._article_container.queue)

    def add_articles(self, articles):
        for article in articles:
            self._article_container.put(article)



from unittest import TestCase
from threading import Thread
from datetime import datetime

from nscrap.runner import ScraperRunner
from nscrap.messenger import Messenger
from nscrap.scraper import Article
from nscrap.keywords import Keyword


class NoneMesseneger(Messenger):

    def send(self, content):
        pass


class RunnerTestCase(TestCase):

    def setUp(self):
        self.messenger = NoneMesseneger()
        self.runner = ScraperRunner(self.messenger)

    def test_shared_article(self):
        number_of_articles = 30
        number_of_threads = 30

        articles = [
            Article(str(index), str(index), datetime.now())
            for index in range(number_of_articles)
        ]

        threads = [
            Thread(target=self.runner.article_container.add_articles, args=(articles,))
            for _ in range(number_of_threads)
        ]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(
            len(self.runner.article_container.get_all_articles()),
            number_of_articles * number_of_threads
        )

    def test_get_new_articles(self):
        existing_articles = [
            Article("1", "1", datetime.now()),
            Article("2", "2", datetime.now()),
            Article("3", "3", datetime.now()),
        ]
        new_articles = [
            Article("4", "4", datetime.now()),
            Article("5", "5", datetime.now()),
        ]
        scraped_articles = existing_articles + new_articles
        self.runner.article_container.add_articles(existing_articles)
        self.assertEqual(self.runner.get_new_articles(scraped_articles), new_articles)

    def test_is_contains_any_keyword(self):
        keywords = [
            Keyword("a"),
            Keyword("b"),
        ]
        self.runner.add_keyword(keywords)
        self.assertTrue(self.runner.is_contains_any_keyword("a"))
        self.assertTrue(self.runner.is_contains_any_keyword("A"))
        self.assertTrue(self.runner.is_contains_any_keyword("b"))
        self.assertFalse(self.runner.is_contains_any_keyword("c"))



from typing import List


class Keyword:

    def __init__(self, keyword: str):
        self.keyword = _sanitize(keyword)

    def is_in(self, word: str) -> bool:
        return self.keyword in word.upper()


def _sanitize(keyword):
    if not keyword:
        raise ValueError("Keyword can not be empty")
    return keyword.strip().upper()


def drop_duplicated_keywords(keywords: List[Keyword]):
    return list(set(keywords))



import setuptools


with open("README.md", "r", encoding="utf8") as file:
    long_description = file.read()


setuptools.setup(
    name="nscrap",
    version="0.0.1",
    author="pparkddo",
    author_email="ehdud3453@gmail.com",
    description="Articles notification by scheduled web scrapers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pparkddo/nscrap",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)
